import ast
import inspect
import linecache
import os
import sys
from time import sleep

import asttokens
import pytest
from littleutils import SimpleNamespace

import executing.executing
from executing import Source, NotOneValueFound
from executing._exceptions import KnownIssue
from executing.executing import is_ipython_cell_code, attr_names_match, is_rewritten_by_pytest

sys.path.append(os.path.dirname(os.path.dirname(__file__)))


def test_pytest():
    from tests.utils import tester

    lst = [1, 2, 3]
    lst2 = tester(lst)
    assert lst == lst2
    lst3 = tester(lst + [4])
    assert (
            [1, 2, 3, 4]
            == lst3
    ), 'message'
    x = tester.x
    assert x is tester


def test_ipython_cell_code():
    assert is_ipython_cell_code(
        SimpleNamespace(
            co_name="<cell line: 1>",
            co_filename="tmp/ipykernel_3/foo",
        )
    )

    assert not is_ipython_cell_code(
        SimpleNamespace(
            co_name="<cell line: 1",
            co_filename="tmp/ipykernel_3/foo",
        )
    )

    assert not is_ipython_cell_code(
        SimpleNamespace(
            co_name="<cell line: 1>",
            co_filename="tmp/ipykernel_3",
        )
    )


def test_attr_names_match():
    assert attr_names_match("foo", "foo")

    assert not attr_names_match("foo", "_foo")
    assert not attr_names_match("foo", "__foo")
    assert not attr_names_match("_foo", "foo")
    assert not attr_names_match("__foo", "foo")

    assert attr_names_match("__foo", "_Class__foo")
    assert not attr_names_match("_Class__foo", "__foo")
    assert not attr_names_match("__foo", "Class__foo")
    assert not attr_names_match("__foo", "_Class_foo")


def test_source_file_text_change(tmpdir):
    # Check that Source.for_filename notices changes in file contents
    # (assuming that linecache can notice)

    path = str(tmpdir.join('foo.py'))
    with open(path, "w") as f:
        f.write("1\n")

    # Initial sanity check.
    source = Source.for_filename(path)
    assert source.text == "1\n"

    # Wait a little before changing the file so that the mtime is different
    # so that linecache.checkcache() notices.
    sleep(0.01)
    with open(path, "w") as f:
        f.write("2\n")
    source = Source.for_filename(path)
    assert source.text == "2\n"


def test_manual_linecache():
    # Test that manually putting lines in linecache
    # under fake filenames works, and the linecache entries aren't removed.
    check_manual_linecache(os.path.join("fake", "path", "to", "foo.py"))
    check_manual_linecache("<my_custom_filename>")


def check_manual_linecache(filename):
    text = "foo\nbar\n"
    lines = text.splitlines(True)
    assert lines == ["foo\n", "bar\n"]

    entry = (len(text), 0, lines, filename)
    linecache.cache[filename] = entry

    # sanity check
    assert linecache.cache[filename] == entry
    assert linecache.getlines(filename) == lines

    # checkcache normally removes the entry because the filename doesn't exist
    linecache.checkcache(filename)
    assert filename not in linecache.cache
    assert linecache.getlines(filename) == []

    # Source.for_filename uses checkcache but makes sure the entry isn't removed
    linecache.cache[filename] = entry
    source = Source.for_filename(filename)
    assert linecache.cache[filename] == entry
    assert source.text == text


def test_exception_catching():
    frame = inspect.currentframe()

    executing.executing.TESTING = True  # this is already the case in all other tests
    # Sanity check that this operation usually raises an exception.
    # This actually depends on executing not working in the presence of pytest.
    with pytest.raises((NotOneValueFound, KnownIssue)):
        assert Source.executing(frame).node is None

    # By contrast, TESTING is usually false when executing is used in real code.
    # In that case, the exception is caught and the result is None.
    executing.executing.TESTING = False
    try:
        assert Source.executing(frame).node is None
    finally:
        executing.executing.TESTING = True


def test_bad_linecache():
    # Test graceful failure when linecache contains source lines that don't match
    # the real code being executed.
    fake_text = "foo bar baz"  # invalid syntax, so source.tree is None
    text = """
frame = inspect.currentframe()
ex = Source.executing(frame)
"""
    filename = "<test_bad_linecache>"
    code = compile(text, filename, "exec")
    linecache.cache[filename] = (len(fake_text), 0, fake_text.splitlines(True), filename)
    globs = dict(globals())
    exec(code, globs)
    ex = globs["ex"]
    frame = globs["frame"]
    assert ex.node is None
    assert ex.statements is None
    assert ex.decorator is None
    assert ex.frame is frame
    assert ex.source.tree is None
    assert ex.source.text == fake_text


if sys.version_info >= (3, 11):
    from executing._position_node_finder import mangled_name
    from textwrap import indent
    import dis

    def test_mangled_name():
        def result(*code_levels):
            code = ""
            for i, level in enumerate(code_levels):
                code += indent(level, "    " * i) + "\n"

            tree = ast.parse(code)

            for parent in ast.walk(tree):
                for child in ast.iter_child_nodes(parent):
                    child.parent = parent

            tree_names = {
                mangled_name(n)
                for n in ast.walk(tree)
                if isinstance(
                    n,
                    (
                        ast.Name,
                        ast.Attribute,
                        ast.alias,
                        ast.FunctionDef,
                        ast.ClassDef,
                        ast.AsyncFunctionDef,
                        ast.ExceptHandler,
                    ),
                )
            }

            def collect_names(code):
                for instruction in dis.get_instructions(code):
                    if instruction.opname in (
                        "STORE_NAME",
                        "LOAD_NAME",
                        "LOAD_GLOBAL",
                        "STORE_FAST",
                        "LOAD_FAST",
                        "LOAD_ATTR",
                        "STORE_ATTR",
                    ):
                        # TODO: "IMPORT_FROM" gets also mangled but is currently not handled by executing
                        #
                        # class Test:
                        #     from __mangle11c.__suc11c import __submodule11c as __subc11
                        # IMPORT_FROM(_Test__submodule11c)
                        # STORE_NAME(_Test__subc11)

                        name = instruction.argval
                        if name in ("__module__", "__qualname__", "__name__"):
                            continue

                        yield name

                for const in code.co_consts:
                    if isinstance(const, type(code)):
                        for name in collect_names(const):
                            yield name

            code_names = set(collect_names(compile(tree, "<code>", "exec")))

            assert code_names == tree_names

            return tree_names

        code = "from __mangle11c.__suc11c import __submodule11c as __subc11"

        assert result(code) == {"__subc11"}

        assert result("class Test:", code) == {"Test", "_Test__subc11"}

        assert result("class Test:", "def func():", code) == {
            "Test",
            "func",
            "_Test__subc11",
        }

        code = """
import __mangled1.submodule1
import __mangled2.__submodule2
import __mangled3.submodule3 as __sub3
import __mangled4.__submodule4 as __sub4
import __mangled5.__submodule5 as sub5
from __mangled6 import submodule6
from __mangle7.sub7 import submodule7
from __mangle8.__sub8 import submodule8
from __mangle9 import submodule9
from __mangle10.sub10 import submodule10
from __mangle11.__sub11 import submodule11
from __mangled6b import __submodule6b
from __mangle7b.sub7b import __submodule7b
from __mangle8b.__sub8b import __submodule8b
from __mangle9b import __submodule9b
from __mangle10b.sub10b import __submodule10b
from __mangle11b.__sub11b import __submodule11b
from __mangled6c import __submodule6c as __subc6
from __mangle7c.suc7c import __submodule7c as __subc7
from __mangle8c.__suc8c import __submodule8c as __subc8
from __mangle9c import __submodule9c as __subc9
from __mangle10c.suc10c import __submodule10c as __subc10
from __mangle11c.__suc11c import __submodule11c as __subc11
"""

        assert result("class Test:", "def func():", code) == {
            "Test",
            "_Test__mangled1",
            "_Test__mangled2",
            "_Test__sub3",
            "_Test__sub4",
            "_Test__subc10",
            "_Test__subc11",
            "_Test__subc6",
            "_Test__subc7",
            "_Test__subc8",
            "_Test__subc9",
            "_Test__submodule10b",
            "_Test__submodule11b",
            "_Test__submodule6b",
            "_Test__submodule7b",
            "_Test__submodule8b",
            "_Test__submodule9b",
            "func",
            "sub5",
            "submodule10",
            "submodule11",
            "submodule6",
            "submodule7",
            "submodule8",
            "submodule9",
        }

        code = """
__mangled_var=3
__not_mangled__=5
normal_var=6

q.__attribute

try:
    pass
except TypeError as __exception:
    pass

for __var in [1]:
    pass
"""
        assert result("class Test:", "def func():", code) == {
            "Test",
            "TypeError",
            "_Test__attribute",
            "_Test__exception",
            "_Test__mangled_var",
            "_Test__var",
            "__not_mangled__",
            "func",
            "normal_var",
            "q",
        }

        # different context

        assert result("class Test:", "def func():", "e.__a=5") == {
            "Test",
            "func",
            "_Test__a",
            "e",
        }

        assert result("class __Test:", "def func():", "e.__a=5") == {
            "__Test",
            "func",
            "_Test__a",
            "e",
        }

        assert result("class __Test:", "e.__a=5") == {
            "__Test",
            "_Test__a",
            "e",
        }

        assert result("class __Test_:", "def func():", "e.__a=5") == {
            "__Test_",
            "func",
            "_Test___a",
            "e",
        }

        assert result("class ___Test_:", "def func():", "e.__a=5") == {
            "___Test_",
            "func",
            "_Test___a",
            "e",
        }


        assert result("class __Testa:","class __Testb:" ,"e.__a=5") == {
            "__Testa",
            "_Testa__Testb",
            "_Testb__a",
            "e",
        }

        assert result(
            "class Test:",
            "def foo(self):",
            "class Patched(self.__attr):",
            "pass",
        ) == {"Patched", "foo", "self", "Test", "_Test__attr"}

        assert result(
            "class _:",
            "def a(self):",
            "self.__thing",
        ) == {"_","a", "self", "__thing"}

        assert result(
            "class __:",
            "def a(self):",
            "self.__thing",
        ) == {"__","a", "self", "__thing"}

        assert result(
            "class Test:",
            "class _:",
            "def a(self):",
            "self.__thing",
        ) == {"Test","_","a", "self", "__thing"}


def test_pytest_rewrite():
    frame = inspect.currentframe()

    # check for assert statements rewrite caused by this assert
    assert is_rewritten_by_pytest(frame.f_code)


def test_no_pytest_rewrite():
    frame=inspect.currentframe()

    # no assert -> no rewrite
    if is_rewritten_by_pytest(frame.f_code):
        raise AssertionError("unexpected pytest assert rewrite")


def test_no_pytest_rewrite_with_consts():
    frame = inspect.currentframe()

    # LOAD_CONST "@py_assert..." should not trigger a false positive
    a = "@py_assert..."

    # no assert -> no rewrite
    if is_rewritten_by_pytest(frame.f_code):
        raise AssertionError("unexpected pytest assert rewrite")


def test_asttext():
    source = Source("file.py", ["a", "b"])
    assert source._asttext is None
    assert source._asttokens is None
    atext = source.asttext()
    assert atext is source.asttext() is source._asttext is not None
    assert source._asttokens is None
    atokens = source.asttokens()
    assert atext.asttokens is atokens is source.asttokens() is source._asttokens is not None



import ast
import dis
from types import CodeType, FrameType
from typing import Any, Callable, Iterator, Optional, Sequence, Set, Tuple, Type, Union, cast
from .executing import EnhancedAST, NotOneValueFound, Source, only, function_node_types, assert_
from ._exceptions import KnownIssue, VerifierFailure

from functools import lru_cache

# the code in this module can use all python>=3.11 features


def parents(node: EnhancedAST) -> Iterator[EnhancedAST]:
    while True:
        if hasattr(node, "parent"):
            node = node.parent
            yield node
        else:
            break

def node_and_parents(node: EnhancedAST) -> Iterator[EnhancedAST]:
    yield node
    yield from parents(node)


def mangled_name(node: EnhancedAST) -> str:
    """

    Parameters:
        node: the node which should be mangled
        name: the name of the node

    Returns:
        The mangled name of `node`
    """
    if isinstance(node, ast.Attribute):
        name = node.attr
    elif isinstance(node, ast.Name):
        name = node.id
    elif isinstance(node, (ast.alias)):
        name = node.asname or node.name.split(".")[0]
    elif isinstance(node, (ast.FunctionDef, ast.ClassDef, ast.AsyncFunctionDef)):
        name = node.name
    elif isinstance(node, ast.ExceptHandler):
        name = node.name or "exc"
    else:
        raise TypeError("no node to mangle")

    if name.startswith("__") and not name.endswith("__"):

        parent,child=node.parent,node

        while not (isinstance(parent,ast.ClassDef) and child not in parent.bases):
            if not hasattr(parent,"parent"):
                break
            parent,child=parent.parent,parent
        else:
            class_name=parent.name.lstrip("_")
            if class_name!="":
                return "_" + class_name + name

            

    return name


@lru_cache(128)
def get_instructions(code: CodeType) -> list[dis.Instruction]:
    return list(dis.get_instructions(code, show_caches=True))


types_cmp_issue_fix = (
    ast.IfExp,
    ast.If,
    ast.Assert,
    ast.While,
)

types_cmp_issue = types_cmp_issue_fix + (
    ast.ListComp,
    ast.SetComp,
    ast.DictComp,
    ast.GeneratorExp,
)

op_type_map = {
    "**": ast.Pow,
    "*": ast.Mult,
    "@": ast.MatMult,
    "//": ast.FloorDiv,
    "/": ast.Div,
    "%": ast.Mod,
    "+": ast.Add,
    "-": ast.Sub,
    "<<": ast.LShift,
    ">>": ast.RShift,
    "&": ast.BitAnd,
    "^": ast.BitXor,
    "|": ast.BitOr,
}


class PositionNodeFinder(object):
    """
    Mapping bytecode to ast-node based on the source positions, which where introduced in pyhon 3.11.
    In general every ast-node can be exactly referenced by its begin/end line/col_offset, which is stored in the bytecode.
    There are only some exceptions for methods and attributes.
    """

    def __init__(self, frame: FrameType, stmts: Set[EnhancedAST], tree: ast.Module, lasti: int, source: Source):
        self.bc_list = get_instructions(frame.f_code)

        self.source = source
        self.decorator: Optional[EnhancedAST] = None

        # work around for https://github.com/python/cpython/issues/96970
        while self.opname(lasti) == "CACHE":
            lasti -= 2

        try:
            # try to map with all match_positions
            self.result = self.find_node(lasti)
        except NotOneValueFound:
            typ: tuple[Type]
            # LOAD_METHOD could load "".join for long "..."%(...) BinOps
            # this can only be associated by using all positions
            if self.opname(lasti) in (
                "LOAD_METHOD",
                "LOAD_ATTR",
                "STORE_ATTR",
                "DELETE_ATTR",
            ):
                # lineno and col_offset of LOAD_METHOD and *_ATTR instructions get set to the beginning of
                # the attribute by the python compiler to improved error messages (PEP-657)
                # we ignore here the start position and try to find the ast-node just by end position and expected node type
                # This is save, because there can only be one attribute ending at a specific point in the source code.
                typ = (ast.Attribute,)
            elif self.opname(lasti) == "CALL":
                # A CALL instruction can be a method call, in which case the lineno and col_offset gets changed by the compiler.
                # Therefore we ignoring here this attributes and searchnig for a Call-node only by end_col_offset and end_lineno.
                # This is save, because there can only be one method ending at a specific point in the source code.
                # One closing ) only belongs to one method.
                typ = (ast.Call,)
            else:
                raise

            self.result = self.find_node(
                lasti,
                match_positions=("end_col_offset", "end_lineno"),
                typ=typ,
            )

        self.known_issues(self.result, self.instruction(lasti))

        self.test_for_decorator(self.result, lasti)

        if self.decorator is None:
            self.verify(self.result, self.instruction(lasti))

    def test_for_decorator(self, node: EnhancedAST, index: int) -> None:
        if (
            isinstance(node.parent, (ast.ClassDef, function_node_types))
            and node in node.parent.decorator_list # type: ignore[attr-defined]
        ):
            node_func = node.parent

            while True:
                # the generated bytecode looks like follow:

                # index    opname
                # ------------------
                # index-4  PRECALL
                # index-2  CACHE
                # index    CALL        <- the call instruction
                # ...      CACHE       some CACHE instructions

                # maybe multiple other bytecode blocks for other decorators
                # index-4  PRECALL
                # index-2  CACHE
                # index    CALL        <- index of the next loop
                # ...      CACHE       some CACHE instructions

                # index+x  STORE_*     the ast-node of this instruction points to the decorated thing

                if self.opname(index - 4) != "PRECALL" or self.opname(index) != "CALL":
                    break

                index += 2

                while self.opname(index) in ("CACHE", "EXTENDED_ARG"):
                    index += 2

                if (
                    self.opname(index).startswith("STORE_")
                    and self.find_node(index) == node_func
                ):
                    self.result = node_func
                    self.decorator = node
                    return

                index += 4

    def known_issues(self, node: EnhancedAST, instruction: dis.Instruction) -> None:
        if instruction.opname in ("COMPARE_OP", "IS_OP", "CONTAINS_OP") and isinstance(
            node, types_cmp_issue
        ):
            if isinstance(node, types_cmp_issue_fix):
                # this is a workaround for https://github.com/python/cpython/issues/95921
                # we can fix cases with only on comparison inside the test condition
                #
                # we can not fix cases like:
                # if a<b<c and d<e<f: pass
                # if (a<b<c)!=d!=e: pass
                # because we don't know which comparison caused the problem

                comparisons = [
                    n
                    for n in ast.walk(node.test) # type: ignore[attr-defined]
                    if isinstance(n, ast.Compare) and len(n.ops) > 1
                ]

                assert_(comparisons, "expected at least one comparison")

                if len(comparisons) == 1:
                    node = self.result = cast(EnhancedAST, comparisons[0])
                else:
                    raise KnownIssue(
                        "multiple chain comparison inside %s can not be fixed" % (node)
                    )

            else:
                # Comprehension and generators get not fixed for now.
                raise KnownIssue("chain comparison inside %s can not be fixed" % (node))

        if isinstance(node, ast.Assert):
            # pytest assigns the position of the assertion to all expressions of the rewritten assertion.
            # All the rewritten expressions get mapped to ast.Assert, which is the wrong ast-node.
            # We don't report this wrong result.
            raise KnownIssue("assert")

        if any(isinstance(n, ast.pattern) for n in node_and_parents(node)):
            # TODO: investigate
            raise KnownIssue("pattern matching ranges seems to be wrong")

        if instruction.opname == "STORE_NAME" and instruction.argval == "__classcell__":
            # handle stores to __classcell__ as KnownIssue,
            # because they get complicated if they are used in `if` or `for` loops
            # example:
            #
            # class X:
            #     # ... something
            #     if some_condition:
            #         def method(self):
            #             super()
            #
            # The `STORE_NAME` instruction gets mapped to the `ast.If` node,
            # because it is the last element in the class.
            # This last element could be anything and gets dificult to verify.

            raise KnownIssue("store __classcell__")

    @staticmethod
    def is_except_cleanup(inst: dis.Instruction, node: EnhancedAST) -> bool:
        if inst.opname not in (
            "STORE_NAME",
            "STORE_FAST",
            "STORE_DEREF",
            "STORE_GLOBAL",
            "DELETE_NAME",
            "DELETE_FAST",
            "DELETE_GLOBAL",
        ):
            return False

        # This bytecode does something exception cleanup related.
        # The position of the instruciton seems to be something in the last ast-node of the ExceptHandler
        # this could be a bug, but it might not be observable in normal python code.

        # example:
        # except Exception as exc:
        #     enum_member._value_ = value

        # other example:
        # STORE_FAST of e was mapped to Constant(value=False)
        # except OSError as e:
        #     if not _ignore_error(e):
        #         raise
        #     return False

        # STORE_FAST of msg was mapped to print(...)
        #  except TypeError as msg:
        #      print("Sorry:", msg, file=file)

        return any(
            isinstance(n, ast.ExceptHandler) and mangled_name(n) == inst.argval
            for n in parents(node)
        )

    def verify(self, node: EnhancedAST, instruction: dis.Instruction) -> None:
        """
        checks if this node could gererate this instruction
        """

        op_name = instruction.opname
        extra_filter: Callable[[EnhancedAST], bool] = lambda e: True
        ctx: Type = type(None)

        def inst_match(opnames: Union[str, Sequence[str]], **kwargs: Any) -> bool:
            """
            match instruction

            Parameters:
                opnames: (str|Seq[str]): inst.opname has to be equal to or in `opname`
                **kwargs: every arg has to match inst.arg

            Returns:
                True if all conditions match the instruction

            """

            if isinstance(opnames, str):
                opnames = [opnames]
            return instruction.opname in opnames and kwargs == {
                k: getattr(instruction, k) for k in kwargs
            }

        def node_match(node_type: Union[Type, Tuple[Type, ...]], **kwargs: Any) -> bool:
            """
            match the ast-node

            Parameters:
                node_type: type of the node
                **kwargs: every `arg` has to be equal `node.arg`
                        or `node.arg` has to be an instance of `arg` if it is a type.
            """
            return isinstance(node, node_type) and all(
                isinstance(getattr(node, k), v)
                if isinstance(v, type)
                else getattr(node, k) == v
                for k, v in kwargs.items()
            )

        if op_name == "CACHE":
            return

        if inst_match("CALL") and node_match((ast.With, ast.AsyncWith)):
            # call to context.__exit__
            return

        if inst_match(("CALL", "LOAD_FAST")) and node_match(
            (ast.ListComp, ast.GeneratorExp, ast.SetComp, ast.DictComp)
        ):
            # call to the generator function
            return

        if inst_match(("CALL", "CALL_FUNCTION_EX")) and node_match(
            (ast.ClassDef, ast.Call)
        ):
            return

        if inst_match(("COMPARE_OP", "IS_OP", "CONTAINS_OP")) and node_match(
            ast.Compare
        ):
            return

        if inst_match("LOAD_NAME", argval="__annotations__") and node_match(
            ast.AnnAssign
        ):
            return

        if (
            (
                inst_match("LOAD_METHOD", argval="join")
                or inst_match(("CALL", "BUILD_STRING"))
            )
            and node_match(ast.BinOp, left=ast.Constant, op=ast.Mod)
            and isinstance(cast(ast.Constant, cast(ast.BinOp, node).left).value, str)
        ):
            # "..."%(...) uses "".join
            return

        if inst_match("STORE_SUBSCR") and node_match(ast.AnnAssign):
            # data: int
            return

        if self.is_except_cleanup(instruction, node):
            return

        if inst_match(("DELETE_NAME", "DELETE_FAST")) and node_match(
            ast.Name, id=instruction.argval, ctx=ast.Del
        ):
            return

        if inst_match("BUILD_STRING") and (
            node_match(ast.JoinedStr) or node_match(ast.BinOp, op=ast.Mod)
        ):
            return

        if inst_match(("BEFORE_WITH","WITH_EXCEPT_START")) and node_match(ast.With):
            return

        if inst_match(("STORE_NAME", "STORE_GLOBAL"), argval="__doc__") and node_match(
            ast.Constant
        ):
            # store docstrings
            return

        if (
            inst_match(("STORE_NAME", "STORE_FAST", "STORE_GLOBAL", "STORE_DEREF"))
            and node_match(ast.ExceptHandler)
            and instruction.argval == mangled_name(node)
        ):
            # store exception in variable
            return

        if (
            inst_match(("STORE_NAME", "STORE_FAST", "STORE_DEREF", "STORE_GLOBAL"))
            and node_match((ast.Import, ast.ImportFrom))
            and any(mangled_name(cast(EnhancedAST, alias)) == instruction.argval for alias in cast(ast.Import, node).names)
        ):
            # store imported module in variable
            return

        if (
            inst_match(("STORE_FAST", "STORE_DEREF", "STORE_NAME", "STORE_GLOBAL"))
            and (
                node_match((ast.FunctionDef, ast.ClassDef, ast.AsyncFunctionDef))
                or node_match(
                    ast.Name,
                    ctx=ast.Store,
                )
            )
            and instruction.argval == mangled_name(node)
        ):
            return

        if False:
            # TODO: match expressions are not supported for now
            if inst_match(("STORE_FAST", "STORE_NAME")) and node_match(
                ast.MatchAs, name=instruction.argval
            ):
                return

            if inst_match("COMPARE_OP", argval="==") and node_match(ast.MatchSequence):
                return

            if inst_match("COMPARE_OP", argval="==") and node_match(ast.MatchValue):
                return

        if inst_match("BINARY_OP") and node_match(
            ast.AugAssign, op=op_type_map[instruction.argrepr.removesuffix("=")]
        ):
            # a+=5
            return

        if node_match(ast.Attribute, ctx=ast.Del) and inst_match(
            "DELETE_ATTR", argval=mangled_name(node)
        ):
            return

        if inst_match(("JUMP_IF_TRUE_OR_POP", "JUMP_IF_FALSE_OR_POP")) and node_match(
            ast.BoolOp
        ):
            # and/or short circuit
            return

        if inst_match("DELETE_SUBSCR") and node_match(ast.Subscript, ctx=ast.Del):
            return

        if node_match(ast.Name, ctx=ast.Load) and inst_match(
            ("LOAD_NAME", "LOAD_FAST", "LOAD_GLOBAL"), argval=mangled_name(node)
        ):
            return

        if node_match(ast.Name, ctx=ast.Del) and inst_match(
            ("DELETE_NAME", "DELETE_GLOBAL"), argval=mangled_name(node)
        ):
            return

        # old verifier

        typ: Type = type(None)
        op_type: Type = type(None)

        if op_name.startswith(("BINARY_SUBSCR", "SLICE+")):
            typ = ast.Subscript
            ctx = ast.Load
        elif op_name.startswith("BINARY_"):
            typ = ast.BinOp
            op_type = op_type_map[instruction.argrepr]
            extra_filter = lambda e: isinstance(cast(ast.BinOp, e).op, op_type)
        elif op_name.startswith("UNARY_"):
            typ = ast.UnaryOp
            op_type = dict(
                UNARY_POSITIVE=ast.UAdd,
                UNARY_NEGATIVE=ast.USub,
                UNARY_NOT=ast.Not,
                UNARY_INVERT=ast.Invert,
            )[op_name]
            extra_filter = lambda e: isinstance(cast(ast.UnaryOp, e).op, op_type)
        elif op_name in ("LOAD_ATTR", "LOAD_METHOD", "LOOKUP_METHOD"):
            typ = ast.Attribute
            ctx = ast.Load
            extra_filter = lambda e: mangled_name(e) == instruction.argval
        elif op_name in (
            "LOAD_NAME",
            "LOAD_GLOBAL",
            "LOAD_FAST",
            "LOAD_DEREF",
            "LOAD_CLASSDEREF",
        ):
            typ = ast.Name
            ctx = ast.Load
            extra_filter = lambda e: cast(ast.Name, e).id == instruction.argval
        elif op_name in ("COMPARE_OP", "IS_OP", "CONTAINS_OP"):
            typ = ast.Compare
            extra_filter = lambda e: len(cast(ast.Compare, e).ops) == 1
        elif op_name.startswith(("STORE_SLICE", "STORE_SUBSCR")):
            ctx = ast.Store
            typ = ast.Subscript
        elif op_name.startswith("STORE_ATTR"):
            ctx = ast.Store
            typ = ast.Attribute
            extra_filter = lambda e: mangled_name(e) == instruction.argval

        node_ctx = getattr(node, "ctx", None)

        ctx_match = (
            ctx is not type(None)
            or not hasattr(node, "ctx")
            or isinstance(node_ctx, ctx)
        )

        # check for old verifier
        if isinstance(node, typ) and ctx_match and extra_filter(node):
            return

        # generate error

        title = "ast.%s is not created from %s" % (
            type(node).__name__,
            instruction.opname,
        )

        raise VerifierFailure(title, node, instruction)

    def instruction(self, index: int) -> dis.Instruction:
        return self.bc_list[index // 2]

    def opname(self, index: int) -> str:
        return self.instruction(index).opname

    def find_node(
        self,
        index: int,
        match_positions: Sequence[str]=("lineno", "end_lineno", "col_offset", "end_col_offset"),
        typ: tuple[Type, ...]=(ast.expr, ast.stmt, ast.excepthandler, ast.pattern),
    ) -> EnhancedAST:
        position = self.instruction(index).positions
        assert position is not None and position.lineno is not None

        return only(
            cast(EnhancedAST, node)
            for node in self.source._nodes_by_line[position.lineno]
            if isinstance(node, typ)
            if not isinstance(node, ast.Expr)
            # matchvalue.value has the same positions as matchvalue themself, so we exclude ast.MatchValue
            if not isinstance(node, ast.MatchValue)
            if all(
                getattr(position, attr) == getattr(node, attr)
                for attr in match_positions
            )
        )



import ast
import operator


def contains_break(node_or_list):
    "search all child nodes except other loops for a break statement"

    if isinstance(node_or_list, ast.AST):
        childs = ast.iter_child_nodes(node_or_list)
    elif isinstance(node_or_list, list):
        childs = node_or_list
    else:
        raise TypeError(node_or_list)

    for child in childs:
        if isinstance(child, (ast.For, ast.While, ast.AsyncFor)):
            if contains_break(child.orelse):
                return True
        elif isinstance(child, ast.Break):
            return True
        elif contains_break(child):
            return True

    return False



class Deadcode:
    @staticmethod
    def annotate(tree):
        deadcode = Deadcode()

        deadcode.annotate_static_values(tree)
        deadcode.walk_deadcode(tree, False)

    def __init__(self):
        self.future_annotations = False

    operator_map = {
        # binary
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.truediv,
        ast.FloorDiv: operator.floordiv,
        ast.Mod: operator.mod,
        ast.Pow: operator.pow,
        ast.LShift: operator.lshift,
        ast.RShift: operator.rshift,
        ast.BitOr: operator.or_,
        ast.BitXor: operator.xor,
        ast.BitAnd: operator.and_,
        ast.MatMult: operator.matmul,
        # unary
        ast.UAdd: operator.pos,
        ast.USub: operator.neg,
        ast.Not: operator.not_,
        ast.Invert: operator.invert,
    }

    def annotate_static_values(self, node):
        for n in ast.iter_child_nodes(node):
            self.annotate_static_values(n)

        try:
            if isinstance(node, ast.Constant):
                node.__static_value = node.value

            elif isinstance(node, ast.Name) and node.id == "__debug__":
                node.__static_value = True

            elif isinstance(node, ast.UnaryOp):
                try:
                    node.__static_value = self.operator_map[type(node.op)](
                        node.operand.__static_value
                    )
                except Exception:
                    pass

            elif isinstance(node, ast.BinOp):
                try:
                    node.__static_value = self.operator_map[type(node.op)](
                        node.left.__static_value, node.right.__static_value
                    )
                except Exception:
                    pass

            elif isinstance(node, ast.Subscript):
                try:
                    node.__static_value = node.value.__static_value[
                        node.slice.__static_value
                    ]
                except Exception:
                    pass

            elif isinstance(node, ast.IfExp):
                cnd = self.static_cnd(node.test)
                if cnd is True:
                    node.__static_value = node.body.__static_value

                elif cnd is False:
                    node.__static_value = node.orelse.__static_value

            elif isinstance(node, ast.BoolOp) and isinstance(node.op, ast.And):
                if all(self.static_cnd(n) is True for n in node.values):
                    node.__static_value = True

                if any(self.static_cnd(n) is False for n in node.values):
                    node.__static_value = False

            elif isinstance(node, ast.BoolOp) and isinstance(node.op, ast.Or):
                if all(self.static_cnd(n) is False for n in node.values):
                    node.__static_value = False

                if any(self.static_cnd(n) is True for n in node.values):
                    node.__static_value = True

        except AttributeError as e:
            if e.name != "_Deadcode__static_value":
                raise

    def static_cnd(self, node):
        try:
            return bool(node.__static_value)
        except AttributeError:
            return None

    def static_value(self, node, deadcode):
        self.walk_deadcode(node, deadcode)
        return self.static_cnd(node)

    def check_stmts(self, stmts, deadcode):
        """
        used to check the body: of a function, if, ...
        """
        for stmt in stmts:
            stmt.deadcode = deadcode

            if self.walk_deadcode(stmt, deadcode):
                deadcode = True
        return deadcode

    def check_childs(self, childs, deadcode):
        """
        used to check childs: function arguments
        """
        for child in childs:
            self.walk_deadcode(child, deadcode)

    def walk_annotation(self, annotation, deadcode):
        if self.future_annotations:
            deadcode = True
        self.walk_deadcode(annotation, deadcode)

    def walk_deadcode(self, node, deadcode):
        "returns True if this statement will never return"

        # this check is not perfect but better than nothing
        # it tries to prevent a lot of "node without associated Bytecode" errors

        # They were generated test driven.
        # Every case represented here is derived from a error where python performed dead code elimination.

        if node is None:
            return

        if isinstance(node, list):
            for child in node:
                self.walk_deadcode(child, deadcode)
            return

        node.deadcode = deadcode or getattr(node, "deadcode", False)

        if isinstance(node, ast.Module):
            for stmt in node.body:
                if isinstance(stmt, ast.ImportFrom):
                    if stmt.module == "__future__" and any(
                        "annotations" == alias.name for alias in stmt.names
                    ):
                        self.future_annotations = True

            self.check_stmts(node.body, deadcode)
        elif isinstance(node, (ast.With, ast.AsyncWith)):
            self.check_childs(node.items, deadcode)
            self.check_stmts(node.body, deadcode)

        elif isinstance(node, (ast.Return, ast.Break, ast.Continue, ast.Raise)):
            if isinstance(node, ast.Raise):
                self.walk_deadcode(node.exc, deadcode)
                self.walk_deadcode(node.cause, deadcode)

            if isinstance(node, ast.Return):
                self.walk_deadcode(node.value, deadcode)

            deadcode = True

        elif isinstance(node, ast.Assert):
            cnd = self.static_value(node.test, deadcode)

            if cnd is False:
                node.deadcode = deadcode
                self.walk_deadcode(node.msg, deadcode)
                deadcode = True

            elif cnd is True:
                node.deadcode = deadcode
                self.walk_deadcode(node.msg, True)

            else:
                node.deadcode = deadcode
                self.walk_deadcode(node.msg, deadcode)

        elif isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            self.walk_annotation(node.args.posonlyargs, deadcode)
            self.walk_annotation(node.args.args, deadcode)
            self.walk_annotation(node.args.kwonlyargs, deadcode)

            self.walk_annotation(node.args.vararg, deadcode)
            self.walk_annotation(node.args.kwarg, deadcode)

            self.check_childs(node.args.kw_defaults, deadcode)
            self.check_childs(node.args.defaults, deadcode)

            self.walk_annotation(node.returns, deadcode)
            self.check_childs(node.decorator_list, deadcode)

            self.check_stmts(node.body, deadcode)

        elif isinstance(node, ast.ClassDef):
            self.check_childs(node.decorator_list, deadcode)

            self.check_childs(node.bases, deadcode)

            self.check_childs(node.keywords, deadcode)

            self.check_stmts(node.body, deadcode)

        elif isinstance(node, ast.If):

            test_value = self.static_value(node.test, deadcode)

            if_is_dead = self.check_stmts(node.body, deadcode or (test_value is False))
            else_is_dead = self.check_stmts(
                node.orelse, deadcode or (test_value is True)
            )

            self.walk_deadcode(node.test, deadcode)

            deadcode = if_is_dead and else_is_dead

        elif isinstance(node, ast.Match):
            self.walk_deadcode(node.subject, deadcode)
            for case_ in node.cases:
                case_.deadcode = deadcode
                self.walk_deadcode(case_.pattern, deadcode)
                self.walk_deadcode(case_.guard, deadcode)

            dead_cases = all(
                [self.check_stmts(case_.body, deadcode) for case_ in node.cases]
            )

            if any(
                isinstance(case_.pattern, ast.MatchAs) and case_.pattern.pattern is None
                for case_ in node.cases
            ):
                # case _:
                deadcode = dead_cases

        elif isinstance(node, (ast.For, ast.AsyncFor)):
            self.walk_deadcode(node.target, deadcode)
            self.walk_deadcode(node.iter, deadcode)
            self.check_stmts(node.body, deadcode)

            else_is_dead = self.check_stmts(node.orelse, deadcode)

            if else_is_dead and not contains_break(node.body):
                # for a in l:
                #     something()
                # else:
                #     return None
                # deadcode()
                deadcode = True

        elif isinstance(node, ast.IfExp):

            test_value = self.static_value(node.test, deadcode)

            self.walk_deadcode(
                node.body, deadcode or (test_value is False)
            )

            self.walk_deadcode(
                node.orelse, deadcode or (test_value is True)
            )

        elif isinstance(node, (ast.While)):
            cnd = self.static_value(node.test, deadcode)

            self.check_stmts(node.body, deadcode or cnd is False)
            else_is_dead = self.check_stmts(node.orelse, deadcode or cnd is True)

            if cnd is True and not contains_break(node):
                # while True: ... no break
                deadcode = True

            if else_is_dead and not contains_break(node.body):
                # for a in l:
                #     something()
                # else:
                #     return None
                # deadcode()
                deadcode = True

        elif isinstance(node, (ast.Try, ast.TryStar)):
            try_dead = self.check_stmts(node.body, deadcode)

            for handler in node.handlers:
                handler.deadcode = deadcode
                self.walk_deadcode(handler.type, deadcode)

            handlers_dead = all(
                [self.check_stmts(h.body, deadcode) for h in node.handlers]
            )
            else_dead = self.check_stmts(node.orelse, try_dead)
            final_dead = self.check_stmts(node.finalbody, deadcode)

            deadcode = (handlers_dead and else_dead) or final_dead

        elif isinstance(node, ast.BoolOp) and isinstance(node.op, ast.And):
            dead_op = deadcode
            for v in node.values:
                if self.static_value(v, dead_op) is False:
                    dead_op = True

        elif isinstance(node, ast.BoolOp) and isinstance(node.op, ast.Or):
            dead_op = deadcode
            for v in node.values:
                if self.static_value(v, dead_op) is True:
                    dead_op = True

        else:

            for n in ast.iter_child_nodes(node):
                self.walk_deadcode(n, deadcode)

        return deadcode


def dump_deadcode(node):
    from rich import print as rprint
    from rich.tree import Tree

    def report(node, tree):
        if isinstance(node, (ast.expr_context, ast.operator, ast.unaryop, ast.cmpop)):
            return

        deadcode = getattr(node, "deadcode", None)
        if deadcode is None:
            deadcode = "[red]<undefined>"
        else:
            deadcode = "[red]dead" if deadcode else "[blue]used"

        name = type(node).__name__

        if isinstance(node, ast.Name):
            name += "(%s)" % node.id

        if isinstance(node, ast.Attribute):
            name += "(.%s)" % node.attr

        if hasattr(node, "_Deadcode__static_value"):
            name += " == %r" % getattr(node, "_Deadcode__static_value")

        t = tree.add("%s %s" % (name, deadcode))
        for child in ast.iter_child_nodes(node):
            report(child, t)

    tree = Tree("ast")
    report(node, tree)
    rprint(tree)


def main():
    import sys

    filename = sys.argv[1]
    print(filename)
    with open(filename) as file:
        tree = ast.parse(file.read())

    Deadcode.annotate(tree)

    dump_deadcode(tree)


if __name__ == "__main__":
    main()
